/**********************************************************************************************
 *
 * This file implements the binomial transformer for computing per-set locality.
 * r' = r.B
 * B is a generalized stochastic Binomial matrix.
 * Ref: Equation 3, Case 1 (Section 3.1) in the paper.
 *
 * Author: Rathijit Sen 
 *         University of Wisconsin-Madison
 *         rathijit@cs.wisc.edu
 *
 **********************************************************************************************/

#include <math.h>
#include "binomial.h"

#define min(x,y) (((x)<=(y))?(x):(y))

per_set_locality_binomial_t::per_set_locality_binomial_t()
{
// Maybe some precomputation can be done here to speedup the transformation in the function below.
}

// The function assumes that num_diff_set_bits>=0, i.e. S<=S'.
// It also assumes that the destination array, rB, is cleared to 0 before this function is called.
// The array rB will have the result of the computation r.B when the function returns
void per_set_locality_binomial_t::transform_r(double r[],double rB[],unsigned int r_size,unsigned int rB_size,unsigned int num_diff_set_bits) const {
    unsigned int k,i;
    if(rB_size>r_size)
        rB_size=r_size;
    if(!num_diff_set_bits) {                            // identity matrix, no transformation required
        for(k=0;k<rB_size;k++)
            rB[k]=r[k];
        return;
    }
    double s_ratio=1.0/(1<<num_diff_set_bits);          // computes S/S'
    long double t1=1.0;
    double t2=s_ratio/(1-s_ratio);
    long double t3;
    for(i=0;i<r_size;i++) {                             // i^th row
        rB[0]+=r[i]*t1;
        t3=1.0;
        for(k=1;k<min(rB_size,i+1);k++) {               // k^th column
            t3*=(t2*(i-k+1))/k;                         // t3 computes iCk * (s_ratio/(1-s_ratio))^k 
            rB[k]+=r[i]*t1*t3;                          // t1*t3 computes iCk * s_ratio^k * (1-s_ratio)^(i-k) 
        }
        t1*=(1-s_ratio);                                // t1 computes (1-s_ratio)^i for the next iteration 
    }
}

