/**********************************************************************************************
 *
 * This file implements the Poisson (approximation to binomial) transformer for computing per-set locality.
 * r' = r.BP
 * BP is a Possion approximation to a generalized stochastic Binomial matrix B.
 * Ref: Appendix B in the paper.
 *
 * Author: Rathijit Sen 
 *         University of Wisconsin-Madison
 *         rathijit@cs.wisc.edu
 *
 **********************************************************************************************/

#include <stdio.h>
#include <math.h>
#include "poisson.h"

// CAUTION: This transformer has been implemented to work with MAX_DIFF_SET_BITS<=8 (r(2^10)-->r(2^18)) and r_size<=512
// For a wider range, you may need to declare precomputed_exp_inc, s3 and s2 as "long double"

per_set_locality_poisson_t::per_set_locality_poisson_t()
 {
    int i;
    for(i=1;i<=MAX_R_TGT_SIZE;i++)
        precomputed_v[i]=1.0/i;        
    for(i=1;i<=MAX_DIFF_SET_BITS;i++) 
        precomputed_exp_inc[i]=exp(-1.0/(1<<i));
}

// The function assumes that num_diff_set_bits>=0, i.e. S<=S'.
// It also assumes that the destination array, rBP, is cleared to 0 before this function is called.
// The array rBP will have the result of the computation r.BP when the function returns
void per_set_locality_poisson_t::transform_r(double r[],double rBP[],unsigned int r_size,unsigned int rBP_size,unsigned int num_diff_set_bits) const {
    unsigned int k,i;
    if(rBP_size>r_size)
        rBP_size=r_size;
    if(!num_diff_set_bits) {                                       // no transformation required as S'=S
        for(k=0;k<rBP_size;k++)
            rBP[k]=r[k];
        return;
    }
    double s_ratio=1.0/(1<<num_diff_set_bits);                     // computes S/S'
    double s3=precomputed_exp_inc[num_diff_set_bits];              // index using the difference in the number of index bits
    double s2=1.0;
    double base_lambda=s_ratio;                                    // determines increment value for lambda 
    double lambda=0;
    for(i=0;i<r_size;i++) {                                        // i^th row
        double s1=s2;
        for(k=0;k<rBP_size;k++,s1*=lambda*precomputed_v[k]) {      // k^th column 
            rBP[k]+=r[i]*s1;
        }
        s2*=s3;                                                    // computes e^(-lambda) for the next iteration
        lambda+=base_lambda;                                       // compute lambda for the next iteration
    }
}

