/**********************************************************************************************
 *
 * This file implements the base class for the miss-ratio predictors.
 * It includes methods used by predictors for all replacement policies.
 *
 * Author: Rathijit Sen 
 *         University of Wisconsin-Madison
 *         rathijit@cs.wisc.edu
 *
 **********************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <assert.h>
#include "util.h"
#include "poisson.h"
#include "pred_mr.h"

pred_mr_t::pred_mr_t() {
    unsigned int m1,m2,set_bits;
    m1=m2=32;
    // see Table 1 for min. and max. associativities for each set size
    for(set_bits=MIN_INDEX_BITS;set_bits<=MAX_INDEX_BITS;set_bits++) {
        minA[set_bits-MIN_INDEX_BITS]=m1;
        maxA[set_bits-MIN_INDEX_BITS]=m2;
        if(set_bits<14)
            m1/=2;
        else if(set_bits>=14)
            m2/=2;
    }
    reset();
}

void pred_mr_t::reset() {
    int i,j;
    for(i=0;i<MAX_R_SRC_SIZE;i++)
        r_src[i]=0;
    for(j=MIN_INDEX_BITS;j<=MAX_INDEX_BITS;j++)
        for(i=0;i<MAX_R_TGT_SIZE;i++) {
            r_tgt[j-MIN_INDEX_BITS][i]=0;
    }
}

// prints the estimated miss ratios
void pred_mr_t::print_miss_ratio() {
    unsigned int set_bits, assoc;
    unsigned int size;  // in MB

    for(size=2;size<=32;size*=2) {
        for(assoc=2;assoc<=32;assoc*=2) {
            set_bits=(unsigned int)(log2(size)+20-6-log2(assoc));  // 2^set_bits=size*(2^20)/(64*assoc)
            if(set_bits<MIN_INDEX_BITS) continue;
            if(set_bits>MAX_INDEX_BITS) continue;
            printf("%dMB %d-way %.8f\n",size,assoc,estimated_mr[set_bits-MIN_INDEX_BITS][assoc]);
        }
    }
}

void pred_mr_t::compute_miss_ratio() {
    unsigned int set_bits, diff_set_bits, assoc;

    struct timeval start_time, end_time;
    double diff=0;
    int nc=0;
    gettimeofday(&start_time,NULL);

    diff_set_bits=0;
    // compute all predictions
    // in practice, use the measured miss-ratio for the current cache configuration. So you will need all-1 predictions.
    for(set_bits=MIN_INDEX_BITS;set_bits<=MAX_INDEX_BITS;set_bits++) {
        // Per-set locality needs to be computed once for each set size
        per_set_locality.transform_r(r_src,r_tgt[diff_set_bits],actual_r_src_size,2*maxA[diff_set_bits],diff_set_bits);
        for(assoc=minA[diff_set_bits];assoc<=maxA[diff_set_bits];assoc*=2) {
            estimated_mr[diff_set_bits][assoc]=compute_miss_ratio_specific(diff_set_bits,assoc);
            nc++;
        }
        diff_set_bits++;
    }

    gettimeofday(&end_time,NULL);
    diff=(end_time.tv_sec*1e6+end_time.tv_usec-start_time.tv_sec*1e6-start_time.tv_usec)/1e6;
    printf("Avg. Computation Time: %.3f msec\n",1000.0*diff/nc);
}

void pred_mr_t::read_r_src(const char *file_path) {
    actual_r_src_size=read_urd(file_path,r_src,MAX_R_SRC_SIZE);
}

