/**********************************************************************************************
 *
 * This file implements the child predictor class for PLRU.
 * Ref: Section 4.4 in the paper.
 *
 * Author: Rathijit Sen 
 *         University of Wisconsin-Madison
 *         rathijit@cs.wisc.edu
 *
 **********************************************************************************************/

#include <stdio.h>
#include <math.h>
#include "pred_plru.h"

// Helper function to compute DCk . p^k . (1-p)^(D-k)
// This is only used for precomputation in the constructor (see below)
double binomial_compute_pmf(int k, int D,double p)
{
    double result,temp;
    int i;
    result=0;
    temp=1.0;
    if(k>D) return 0;
    if(k==0) return pow((1-p),D);
    for(i=0;i<k;i++)
    {
        temp*=(((double)(D-i))/(k-i));
        temp*=p;
    }
    temp*=pow((1-p),(D-k));
    result+=temp;
    return result;
}

pred_plru_t::pred_plru_t() {
    int i,j;
    for(i=0;i<=(MAX_ASSOC/2)+1;i++)
        for(j=0;j<=i-2;j++)
            plru_factors1[i][j]=binomial_compute_pmf(j,i-2,0.5);

    for(i=4;i<2*MAX_ASSOC;i++)
        for(j=0;j<=i-3;j++)
            plru_factors2[i][j]=binomial_compute_pmf(j,i-3,0.5);
    
    reset();
}

void pred_plru_t::reset() {
    int i,j;
    for(i=0;i<=18;i++)
        for(j=0;j<=32;j++)
            plru_computed[i][j]=false;
    pred_mr_t::reset();
}

// Implements Equation 13
void pred_plru_t::plru_compute_phi_4(unsigned int num_diff_set_bits) {
    unsigned int assoc=4;
    unsigned int C=2;    /* log2(A')=2 */
    if(plru_computed[num_diff_set_bits][assoc]) return;
    plru_computed[num_diff_set_bits][assoc]=true;
    int i;
    double t1=0;
    for(i=0;i<=C;i++)
        plru_phi_4[num_diff_set_bits][i]=1;

    t1=r_tgt[num_diff_set_bits][3];
    
    plru_phi_4[num_diff_set_bits][i]=0.5+0.5*(1-t1)*2/3.0;
    i++;
    for(;i<2*MAX_ASSOC;i++)
        plru_phi_4[num_diff_set_bits][i]=plru_phi_4[num_diff_set_bits][i-1]*(1-plru_phi_4[num_diff_set_bits][3]);
}

// Implements Equation 14
void pred_plru_t::plru_compute_phi_8(unsigned int num_diff_set_bits) {
    unsigned int assoc=8;
    unsigned int C=3;  /* log2(A')=3 */
    if(plru_computed[num_diff_set_bits][assoc]) return;
    plru_computed[num_diff_set_bits][assoc]=true;
    int i,j;
    for(i=0;i<=C;i++)
        plru_phi_8[num_diff_set_bits][i]=1;

    for(;i<=(assoc/2+1);i++) {
        double t1=0;
        for(j=0;j<=i-2;j++)
            t1+=plru_factors1[i][j]*plru_phi_4[num_diff_set_bits][1+j];
        plru_phi_8[num_diff_set_bits][i]=0.5*(plru_phi_8[num_diff_set_bits][i-1]+t1);
    }

    for(;i<2*MAX_ASSOC;i++) {
        double t1=0;
        for(j=0;j<=i-3;j++)
            t1+=plru_factors2[i][j]*plru_phi_4[num_diff_set_bits][2+j];
        plru_phi_8[num_diff_set_bits][i]=0.5*(plru_phi_8[num_diff_set_bits][i-1]+t1);
    }
}

// Implements Equation 14
void pred_plru_t::plru_compute_phi_16(unsigned int num_diff_set_bits) {
    unsigned int assoc=16;
    unsigned int C=4; /* log2(A')=4 */
    if(plru_computed[num_diff_set_bits][assoc]) return;
    plru_computed[num_diff_set_bits][assoc]=true;
    int i,j;
    for(i=0;i<=C;i++)
        plru_phi_16[num_diff_set_bits][i]=1;

    for(;i<=(assoc/2+1);i++) {
        double t1=0;
        for(j=0;j<=i-2;j++)
            t1+=plru_factors1[i][j]*plru_phi_8[num_diff_set_bits][1+j];
        plru_phi_16[num_diff_set_bits][i]=0.5*(plru_phi_16[num_diff_set_bits][i-1]+t1);
    }

    for(;i<2*MAX_ASSOC;i++) {
        double t1=0;
        for(j=0;j<=i-3;j++)
            t1+=plru_factors2[i][j]*plru_phi_8[num_diff_set_bits][2+j];
        plru_phi_16[num_diff_set_bits][i]=0.5*(plru_phi_16[num_diff_set_bits][i-1]+t1);
    }
}

// Implements Equation 14
void pred_plru_t::plru_compute_phi_32(unsigned int num_diff_set_bits) {
    unsigned int assoc=32;
    unsigned int C=5; /* log2(A')=5 */
    if(plru_computed[num_diff_set_bits][assoc]) return;
    plru_computed[num_diff_set_bits][assoc]=true;
    int i,j;
    for(i=0;i<=C;i++)
        plru_phi_32[num_diff_set_bits][i]=1;

    for(;i<=(assoc/2+1);i++) {
        double t1=0;
        for(j=0;j<=i-2;j++)
            t1+=plru_factors1[i][j]*plru_phi_16[num_diff_set_bits][1+j];
        plru_phi_32[num_diff_set_bits][i]=0.5*(plru_phi_32[num_diff_set_bits][i-1]+t1);
    }

    for(;i<2*MAX_ASSOC;i++) {
        double t1=0;
        for(j=0;j<=i-3;j++)
            t1+=plru_factors2[i][j]*plru_phi_16[num_diff_set_bits][2+j];
        plru_phi_32[num_diff_set_bits][i]=0.5*(plru_phi_32[num_diff_set_bits][i-1]+t1);
    }
}

double pred_plru_t::compute_miss_ratio_specific(unsigned int num_diff_set_bits,unsigned int assoc) {
    double p=0;
    int i;
    if(assoc<=2) {
        // \phi(A'=2,PLRU)=\phi(A'=2,LRU)
        for(i=0;i<assoc;i++)
            p+=r_tgt[num_diff_set_bits][i];
        return 1-p;
    }
    const double *ptr;
    switch(assoc) {
        case 4:plru_compute_phi_4(num_diff_set_bits);ptr=&plru_phi_4[num_diff_set_bits][0];break;
        case 8:plru_compute_phi_4(num_diff_set_bits);plru_compute_phi_8(num_diff_set_bits);ptr=&plru_phi_8[num_diff_set_bits][0];break;
        case 16:plru_compute_phi_4(num_diff_set_bits);plru_compute_phi_8(num_diff_set_bits);plru_compute_phi_16(num_diff_set_bits);ptr=&plru_phi_16[num_diff_set_bits][0];break;
        case 32:plru_compute_phi_4(num_diff_set_bits);plru_compute_phi_8(num_diff_set_bits);plru_compute_phi_16(num_diff_set_bits);plru_compute_phi_32(num_diff_set_bits);ptr=&plru_phi_32[num_diff_set_bits][0];break;
        default: ptr=NULL;
    }
    // compute the dot-product r_tgt.\phi(PLRU)
    for(i=0;i<2*assoc;i++) 
        p+=r_tgt[num_diff_set_bits][i]*ptr[i];
    return 1-p;
}

