/**********************************************************************************************
 *
 * This file declares utilities to read r from a file, compute d from r
 * Ref: Equation 1 in the paper.
 *
 * Author: Rathijit Sen 
 *         University of Wisconsin-Madison
 *         rathijit@cs.wisc.edu
 *
 **********************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

// This function reads r from a text file and returns 1 + the max. URD read 

// Each line of the input file for r must be of the form "i P(URD=i)" with i in increasing order. 
// r_i = P(URD=i) must be in [0,1], \sum {r_i} <= 1.
// Refer to the example files example_r_S1.txt and example_r_S1024.txt.
//
// r_\infty is never explicitly specified in the input file. Whenever needed, e.g., estimating d, it is calculated as 1-\sum {r_i}
// So, it is usually the case that for the input file \sum {r_i} < 1.
// r_\infty does not participate in matrix transformations for per-set locality.

unsigned int read_urd(const char *file_path,double r[],unsigned int max_size) {
    FILE *fp=fopen(file_path,"r");
    assert(fp);
    unsigned int urd;
    char buffer[100];
    double prob;
    unsigned int actual_size=0;
    while(!feof(fp)) {
        if(fscanf(fp,"%d %s",&urd,buffer)==EOF)
            break;
        if(urd>=max_size)
            break;                         /* assumes that URDs in the file are in increasing order */
        prob=strtod(buffer,NULL);
        r[urd]=prob;
        actual_size=urd+1;
    }
    fclose(fp);
    return actual_size;
}

// This function implements Equation 1
// It assumes that r[] is available for at least d_size-1 positions
void compute_d(double r[],double d[],unsigned int d_size) {
    unsigned int i,j;
    double temp_sum=0;
    j=0;
    d[0]=0;
    for(i=1;i<d_size;i++) {
        for(;j<i;j++)
            temp_sum+=r[j];
        d[i]=d[i-1]+1.0/(1.0-temp_sum);
    }
}

